#include "Global.h"
#include "Util.h"
#include "Object.h"
#include "Hero.h"

shAttack *UniversalSolventDamage =
    newOtherAttack (shAttack::kNoAttack, 0, kCorrosive, 2, 12);
shAttack *PlasmaDamage =
    newOtherAttack (shAttack::kNoAttack, 0, kElectrical, 4, 24);
shAttack *NapalmDamage =
    newOtherAttack (shAttack::kNoAttack, 0, kBurning, 4, 24);
shAttack *LiquidNitrogenDamage =
    newOtherAttack (shAttack::kNoAttack, 0, kFreezing, 4, 24);
shAttack *PoisonDamage =
    newOtherAttack (shAttack::kNoAttack, 0, kPoisonous, 1, 3);

static void
teleportVat (shFeature *vat)
{
    int newX; int newY;
    int prevroom = Level->getRoomID (Hero.mX, Hero.mY);
    Level->findSuitableStairSquare (&newX, &newY);
    if (!Level->getFeature (newX, newY)) {
        if (!Hero.isBlind ()) {
            I->p ("The vat flickers and disappears!");
        } else {
            I->p ("The vat seems to be gone.");
        }
        Level->addVat (newX, newY);
        shFeature *newvat = Level->getFeature (newX, newY);
        newvat->mVat = vat->mVat; /* Transfer all vat properties. */
        Level->removeFeature (vat);
        if (Hero.canSee (newX, newY)) {
            I->p ("Suddenly a vat materializes out of nowhere and falls down!");
        } /* See if the vat falls on someone. */
        shCreature *c = Level->getCreature (newX, newY);
        if (c) {
            if (c->sufferDamage (FallingVatAttack)) {
                if (Hero.canSee (c)) {
                    I->p ("It crushes %s into a pulp!", AN (c));
                    I->pauseXY (c->mX, c->mY);
                } else if (Hero.canHearThoughts (c)) {
                    I->p ("The mind of %s emanates pure terror for an instant.", AN (c));
                    Hero.checkConcentration (); /* Contact with terrified mind. */
                    I->p ("Then thoughts of %s abruptly vanish.", THE (c));
                    I->pauseXY (c->mX, c->mY);
                }
                c->die (kSlain);
            } else if (Hero.canSee (c)) {
                I->p ("It lands right on %s!", AN (c));
                I->pauseXY (c->mX, c->mY);
                if (c->canSee (&Hero)) {
                    c->newEnemy (&Hero);
                }
            } else if (Hero.canHearThoughts (c)) {
                I->p ("The mind of %s emanates fear for a moment.", AN (c));
                I->pauseXY (c->mX, c->mY);
                if (c->hasTelepathy ()) {
                    I->p("Somehow %s blames you for this.", THE (c));
                    c->newEnemy (&Hero);
                }
            }
        } /* Did the vat belong to a shopkeeper? */
        if (Hero.isInShop ()) {
            int newroom = Level->getRoomID (newX, newY);
            if (prevroom == newroom) {
                Hero.movedVat (); /* Harmless furniture rearrange. */
            } else {
                Hero.stolenVat (); /* Vat landed outside shop. */
            }
        }
    } else {
        I->p ("Nothing seems to happen.");
    }
}

/* Splashes objects on floor with canister contents. */
static void
pourCanisterOnObjects (shObject *can, int x, int y)
{
    int seen = Hero.canSee (x, y);
    shObjectVector *v = Level->getObjects (x, y);
    if (v && can->myIlk ()->mMissileAttack) { /* Is explosive. */
        for (int i = v->count () - 1; i >= 0; --i) {
            shObject *obj = v->get (i);
            if (obj->sufferDamage (can->myIlk ()->mMissileAttack, x, y, 0, 0)) {
                if (seen) {
                    can->setIlkKnown ();
                } else {
                    can->maybeName ();
                }
                v->remove (obj);
                delete obj;
            }
        }
        if (!v->count ()) {
            delete v;
            v = NULL;
            Level->setObjects (x, y, NULL);
        }
        return;
    } /* Try for special interaction. */
    if (v) for (int i = v->count () - 1; i >= 0; --i) {
        shObject *obj = v->get (i);
        switch (can->mIlkId) {
        case kObjBrain:
            if (obj->isA (kObjBioComputer)) {
                if (seen) {
                    I->p ("%s devours the brain!", THE (obj));
                    can->setIlkKnown ();
                } else {
                    I->p ("You hear content munching sounds.");
                    can->maybeName ();
                }
                obj->mEnhancement = mini(+5, obj->mEnhancement + can->mBugginess + 1);
                return;
            } else if (obj->isA (kObjBioArmor)) {
                if (Hero.canSee (x, y)) {
                    I->p ("%s envelops the brain!", THE (obj));
                    can->setIlkKnown ();
                } else {
                    I->p ("You hear a slurping sound.");
                    can->maybeName ();
                }
                obj->mIlkId = kObjBioComputer;
                obj->mEnhancement = RNG (-3, +3) + can->mBugginess;
                return;
            }
            break;
        case kObjMutagenCanister:
            if (obj->isA (kObjBioArmor)) {
                int die = RNG (2);
                if (seen) {
                    I->p ("%s soaks up the liquid!", THE (obj));
                    if (die) {
                        I->p ("It develops a fatal mutation and dies.");
                    } else {
                        I->p ("It develops a harmful mutation.");
                    }
                    can->setIlkKnown ();
                }
                if (die) {
                    v->remove (obj);
                    delete obj;
                    if (!v->count ()) {
                        delete v;
                        v = NULL;
                        Level->setObjects (x, y, NULL);
                    }
                } else {
                    obj->mIlkId = kObjMBioArmor0;
                    obj->mEnhancement = maxi (-(obj->myIlk ()->mMaxEnhancement),
                        obj->mEnhancement - 1);
                }
                return;
            }
            break;
        case kObjGainAbilityCanister:
            if (!obj->isA (kObjBioComputer) && !obj->isA (kObjBioArmor)) {
                break;
            }
            if (seen) {
                I->p ("%s soaks up the liquid!", THE (obj));
            }
            if (obj->isA (kObjBioComputer)) {
                if (can->isBuggy ()) {
                    obj->mEnhancement = -5;
                    if (seen) {
                        I->p ("It loses its ability to process input.");
                        can->setIlkKnown ();
                    }
                } else {
                    if (seen && (obj->mEnhancement == 5 ||
                        (obj->mEnhancement == 4 && !can->isOptimized ())))
                    {
                        I->p("Nothing else seems to happen.");
                        can->maybeName ();
                    } else {
                        if (seen) {
                            I->p ("Its abilities improve!");
                            can->setIlkKnown ();
                        }
                        ++obj->mEnhancement;
                    }
                }
            } else if (can->isBuggy ()) {
                if (obj->mIlkId == kObjBioArmor ||
                    obj->isA (kObjMBioArmor0))
                {
                    if (obj->isEnhancementKnown () && seen) {
                        I->p ("It suffers.");
                        can->setIlkKnown ();
                    }
                    obj->mEnhancement = maxi (-(obj->myIlk ()->mMaxEnhancement),
                        obj->mEnhancement - 1);
                } else {
                    if (seen) {
                        I->p ("It regresses in development!");
                        can->setIlkKnown ();
                    }
                    obj->mIlkId = obj->myIlk ()->mParent;
                }
            } else {
                if (obj->mIlkId == kObjBioArmor) {
                    if (seen) {
                        I->p ("It gains a new ability!");
                        can->setIlkKnown ();
                    }
                    switch (RNG (3)) {
                    case 0: obj->mIlkId = kObjEBioArmor1; break;
                    case 1: obj->mIlkId = kObjEBioArmor2; break;
                    case 2: obj->mIlkId = kObjEBioArmor3; break;
                    }
                } else if (obj->mIlkId == kObjMBioArmor0) {
                    if (seen) {
                        I->p ("Its mutation develops into somewhat beneficial one.");
                        can->setIlkKnown ();
                    }
                    if (RNG (2)) {
                        obj->mIlkId = kObjMBioArmor1;
                    } else {
                        obj->mIlkId = kObjMBioArmor2;
                    }
                } else if (can->isOptimized () &&
                       (obj->mIlkId == kObjEBioArmor1 ||
                        obj->mIlkId == kObjEBioArmor2 ||
                        obj->mIlkId == kObjEBioArmor3))
                {
                    if (seen) {
                        I->p ("It is further enhanced!");
                        can->setIlkKnown ();
                    }
                    switch (obj->mIlkId) {
                    case kObjEBioArmor1:
                        if (RNG (2)) {
                            obj->mIlkId = kObjSBioArmor1;
                        } else {
                            obj->mIlkId = kObjSBioArmor2;
                        }
                        break;
                    case kObjEBioArmor2:
                        obj->mIlkId = kObjSBioArmor3; break;
                    case kObjEBioArmor3:
                        obj->mIlkId = kObjSBioArmor4; break;
                    default: break;
                    }
                } else {
                    if (obj->isEnhancementKnown () && seen) {
                        I->p ("Looks like %s enjoyed it.", THE (obj));
                        can->setIlkKnown ();
                    }
                    obj->mEnhancement = mini (obj->myIlk ()->mMaxEnhancement,
                        obj->mEnhancement + 1);
                }
            }
            return;
        default: /* Has no special interaction. No point in iterating. */
            return;
        }
    } /* If it survived until this point it means the floor is contaminated. */
    if (can->isA (kObjMutagenCanister)) {
        Level->getSquare(x, y)->mFlags |= shSquare::kRadioactive;
    }
}

static int
pourCanisterAtFeature (shObject *can, shFeature *f, int x, int y)
{
    f->mTrapUnknown = 0; /* Better reveal it. */
    switch (f->mType) {
    case shFeature::kAcidPit:
        switch (can->mIlkId) {
        case kObjAntimatter:
            if (Level->getLevelBelow ()) {
                if (!Hero.isBlind ()) {
                    I->p ("The acid is annihilated.");
                }
                f->mType = shFeature::kHole;
                Level->checkTraps (x, y);
            } else {
                f->mType = shFeature::kPit;
            }
            return FULLTURN;
        case kObjWater:
            f->mType = shFeature::kPit;
            if (!Hero.isBlind ()) {
                I->p ("The acid boils violently as it is neutralized.");
                can->setIlkKnown ();
            } else {
                I->p ("You hear boiling.");
                can->maybeName ();
            }
            return FULLTURN;
        default:
            if (!Hero.isBlind ()) {
                I->p ("The acid boils.");
                can->maybeName ();
            } else {
                I->p ("You hear boiling.");
                can->maybeName ();
            }
            return FULLTURN;
        }
    case shFeature::kPit:
        switch (can->mIlkId) {
        case kObjAntimatter:
            if (Level->getLevelBelow ()) {
                f->mType = shFeature::kHole;
                Level->checkTraps (x, y);
            }
            return FULLTURN;
        case kObjUniversalSolvent:
            f->mType = shFeature::kAcidPit;
            if (!Hero.isBlind ()) {
                I->p ("The pit is filled with acid!");
            } else {
                I->p ("You smell acidic odor.");
            }
            can->setIlkKnown ();
            return FULLTURN;
        default:
            I->p ("You pour out the contents into %s.", THE (f));
            pourCanisterOnObjects (can, x, y);
            can->maybeName ();
            return FULLTURN;
        }
    case shFeature::kHole:
        I->p ("You pour out the contents into %s.", THE (f));
        return FULLTURN;
    case shFeature::kRadTrap:
        switch (can->mIlkId) {
        case kObjAntimatter:
            if (!Hero.isBlind ()) {
                I->p ("The rad trap is annihilated!");
            }
            if (Level->getLevelBelow ()) {
                f->mType = shFeature::kHole;
            } else {
                f->mType = shFeature::kPit;
            }
            Level->checkTraps (x, y, 100);
            break;
        case kObjUniversalSolvent:
            if (!Hero.isBlind ()) {
                I->p ("The rad trap dissolves!");
            } else {
                I->p ("You smell acidic odor.");
            }
            Level->removeFeature (f);
            can->setIlkKnown ();
            break;
        default:
            I->p ("You pour out the contents into %s.", THE (f));
            can->maybeName ();
            break;
        }
        return FULLTURN;
    case shFeature::kVat:
        switch (can->mIlkId) {
        case kObjAntimatter:
            if (Hero.canSee (x, y)) {
                I->p ("The vat is annihilated!");
            }
            Level->removeFeature (f);
            can->setIlkKnown ();
            Level->addTrap (x, y, shFeature::kHole);
            if (Hero.isInShop ()) {
                Hero.damagedShop (x, y);
            }
            Level->checkTraps (x, y, 100);
            return FULLTURN;
        case kObjLNO:
            I->p ("The sludge freezes!");
            can->setIlkKnown ();
            return FULLTURN;
        case kObjNapalm: case kObjPlasmaCanister:
            I->p ("The sludge boils!");
            can->maybeName ();
            return FULLTURN;
        case kObjBeer: case kObjNanoCola: case kObjB3: case kObjCoffee:
            I->p ("The sludge fizzes.");
            can->maybeName ();
            return FULLTURN;
        case kObjSuperGlue:
            I->p ("The sludge thickens.");
            can->setIlkKnown ();
            return FULLTURN;
        case kObjUniversalSolvent: case kObjWater:
            I->p ("The sludge thins.");
            can->maybeName ();
            return FULLTURN;
        case kObjMutagenCanister:
            I->p ("The sludge changes color.");
            f->mVat.mHealthy -= 2;
            f->mVat.mRadioactive++;
            can->setIlkKnown ();
            return FULLTURN;
        case kObjPoisonCanister:
            I->p ("The sludge seems less nutritous now.");
            f->mVat.mHealthy -= 2;
            can->setIlkKnown ();
            return FULLTURN;
        case kObjFullHealingCanister:
            f->mVat.mHealthy += 2; /* Fall through. */
        case kObjHealingCanister:
            I->p ("The sludge seems more nutritious now.");
            f->mVat.mHealthy += 1;
            can->maybeName ();
            return FULLTURN;
        case kObjGainAbilityCanister:
            f->mVat.mHealthy += 2; /* Fall through. */
        case kObjRestorationCanister:
            I->p ("You've improved the sludge recipe.");
            f->mVat.mHealthy += 2;
            can->maybeName ();
            return FULLTURN;
        case kObjBrain:
            I->p ("You drain a brain.");
            f->mVat.mHealthy += 2;
            can->setIlkKnown ();
            return FULLTURN;
        case kObjRadAway:
            I->p ("The sludge seems purified.");
            f->mVat.mHealthy += 1;
            f->mVat.mRadioactive = 0;
            can->setIlkKnown ();
            return FULLTURN;
        case kObjSpeedCanister:
            I->p ("The sludge is churning more rapidly now.");
            can->setIlkKnown ();
            return FULLTURN;
        case kObjSpiceMelange:
            teleportVat (f);
            can->setIlkKnown ();
            return FULLTURN;
        case kObjCannedEmbryo: {
            if (!Hero.isBlind ()) {
                I->p ("A larva falls out and drowns in the sludge!");
                can->setIlkKnown ();
            }
            f->mVat.mHealthy -= 3;
            /* Give exp for killing alien. */
            shMonsterIlk *m = MonIlks[kMonFacehugger];
            Hero.earnXP (m->mBaseLevel);
            /* Count wasted embryos too. */
            MonIlks[kMonAlienEmbryo]->mKills += 1;
            if (Hero.isBlind ()) {
                I->p ("You feel more experienced.");
                can->maybeName ();
            }
            return FULLTURN;
        }
        default:
            can->setIlkKnown (); /* This might help nail down the bug. */
            I->p ("You almost breach space-time continuity by triggering unimplemented effect.");
            I->p ("You'd better report this bug and how you caused it.");
            return FULLTURN;
        }
    case shFeature::kPortableHole:
        switch (can->mIlkId) {
        case kObjAntimatter:
            I->p ("The portable hole is annihilated along with %s!", THE (Level->getSquare (x, y)));
            Level->removeFeature (f);
            Level->dig (x, y);
            break;
        case kObjUniversalSolvent:
            if (!Hero.isBlind ()) {
                I->p ("The portable hole dissolves!");
            } else {
                I->p ("You smell acidic odor.");
            }
            Level->removeFeature (f);
            can->setIlkKnown ();
            break;
        default:
            I->p ("You pour out the contents onto %s.", THE (Level->getSquare (x, y)));
        }
        return FULLTURN;
    case shFeature::kDoorOpen:
    default:
        I->p ("You pour out the contents onto %s.", THE (f));
        if (can->myIlk ()->mMissileAttack) {
            Level->areaEffectFeature (can->myIlk ()->mMissileAttack, NULL,
                x, y, &Hero);
        }
        pourCanisterOnObjects (can, x, y);
        can->maybeName ();
        return FULLTURN;
    }

}

static int useUniversalSolvent (shObject *can);

static int
useCommonCanister (shObject *can)
{
    shDirection where = I->getDirection ();
    if (kNoDirection == where) {
        return 0;
    }
    switch (where) {
    case kDown: {
        shFeature *f = Level->getFeature (Hero.mX, Hero.mY);
        if (f) return pourCanisterAtFeature (can, f, Hero.mX, Hero.mY);
        I->p ("You pour out the contents onto %s.",
            THE (Level->getSquare (Hero.mX, Hero.mY)));
        pourCanisterOnObjects (can, Hero.mX, Hero.mY);
        can->maybeName ();
        return FULLTURN;
    }
    case kOrigin:
        if (can->isA (kObjUniversalSolvent)) return useUniversalSolvent (can);
        /* Otherwise fall through. */
    case kUp:
        if (can->myIlk ()->mMissileAttack) {
            can->setIlkKnown ();
            switch (can->mIlkId) {
            case kObjAntimatter: I->p ("You spill antimatter."); break;
            case kObjNapalm: I->p ("It ignites!"); break;
            case kObjLNO: I->p ("It freezes!"); break;
            case kObjPlasmaCanister: I->p ("It shocks you!"); break;
            case kObjUniversalSolvent: I->p ("It is acidic!"); break;
            case kObjSuperGlue: I->p ("This is sticky."); break;
            default: I->p ("You soak yourself."); break;
            }
            if (Hero.sufferDamage (can->myIlk ()->mMissileAttack)) {
                char *buf = GetBuf ();
                sprintf (buf, "spilling a %s %s",
                    can->getShortDescription (),
                    where == kUp ? "upwards" : "at self");
                Hero.die(kKilled, buf);
            }
        }
        return FULLTURN;
    default: {
        int x = Hero.mX, y = Hero.mY;
        if (!Level->moveForward (where, &x, &y)) {
            I->p ("You stick the canister outside the map and pour it out.");
            I->p ("Its contents are harmlessly deallocated.");
            return FULLTURN;
        }
        shCreature *c = Level->getCreature (x, y);
        if (NULL == c) {
            shFeature *f = Level->getFeature (x, y);
            if (f) {
                return pourCanisterAtFeature (can, f, x, y);
            } else {
                I->p ("You pour out the contents onto %s.",
                    THE (Level->getSquare (x, y)));
            }
            if (can->isA (kObjAntimatter) && !Level->isFloor (x, y)) {
                I->p ("%s is annihilated!", THE (Level->getSquare (x, y)));
                Level->dig (x, y);
                return FULLTURN;
            }
            if (can->myIlk ()->mMissileAttack) {
                Level->areaEffectFeature (can->myIlk ()->mMissileAttack, NULL,
                    x, y, &Hero);
            }
            pourCanisterOnObjects (can, x, y);
            return FULLTURN;
        }
        I->p ("You splash the contents of canister at %s.", THE (c));
        /* Instakill because player expends whole canister. Yes, even Shodan. */
        if (can->isA (kObjAntimatter)) {
            I->p ("%s is annihilated!", THE (c));
            c->die (kAnnihilated);
            return FULLTURN;
        }
        if (can->myIlk ()->mMissileAttack) {
            switch (can->mIlkId) { /* TODO: Change to 3rd person. */
            case kObjNapalm: I->p ("You burn %s!", THE (c)); break;
            case kObjLNO: I->p ("You freeze %s!", THE (c)); break;
            case kObjPlasmaCanister: I->p ("You shock %s!", THE (c)); break;
            case kObjUniversalSolvent: I->p ("You dissolve %s!", THE (c)); break;
            case kObjSuperGlue: I->p ("You coat %s with glue.", THE (c)); break;
            default:
                I->p ("You do something evil and buggy to %s.", THE (c)); break;
            }
            if (c->sufferDamage (can->myIlk ()->mMissileAttack)) {
                c->die (kSlain);
            }
            can->setIlkKnown ();
        } else {
            I->p ("You soak %s.", THE (c));
            can->maybeName ();
        }
        return FULLTURN;
    }
    }
    return FULLTURN; /* Not reached. */
}

static int
quaffAntimatter (shObject *can)
{
    I->p ("You drink antimatter.");
    can->setIlkKnown ();
    Hero.die (kAnnihilated, "drinking a canister of antimatter");
    return FULLTURN;
}

static int
quaffBeer (shObject *can)
{
    can->setIlkKnown ();
    if (Hero.is (kFrightened)) {
        Hero.cure (kFrightened);
        I->p ("Liquid courage!");
    } else {
        I->p ("Mmmmm... beer!");
    }
    Hero.inflict (kConfused, NDX (2, 50) * FULLTURN);
    return FULLTURN;
}

static int
quaffNanoCola (shObject *can)
{
    int amt = RNG (2+can->mBugginess, 6);
    can->setIlkKnown ();

    Hero.mPsiDrain -= amt;
    Hero.mAbil.mPsi += amt;
    I->p ("You feel invigorated!");
    return FULLTURN;
}

static int
quaffCoffee (shObject *can)
{
    if (!can->isIlkKnown ()) {
        I->p ("As you open the canister it heats itself up and you smell ... coffee?");
    } else {
        I->p ("You drink %s coffee.", can->isBuggy () ? "lukewarm" : "hot");
    }
    if (!Hero.getTimeOut (CAFFEINE)) {
        Hero.mInnateResistances[kMesmerizing] += 25;
    } /* Values for timeouts are those used by disk of hypnosis in reverse. */
    if (can->isBuggy ()) {
        Hero.setTimeOut (CAFFEINE, RNG (5, 25) * 1000, 0);
    } else if (can->isDebugged ()) {
        Hero.setTimeOut (CAFFEINE, RNG (6, 40) * 1000, 0);
    } else {
        Hero.setTimeOut (CAFFEINE, RNG (12, 80) * 1000, 0);
    }
    can->setIlkKnown ();
    return FULLTURN;
}

static int
quaffSuperGlue (shObject *can)
{
    can->setIlkKnown ();
    I->p ("The canister sticks to your tongue!");
    I->p ("You look ridiculous!");
    Hero.setStoryFlag ("superglued tongue", 1);
    return FULLTURN;
}

static int
quaffUniversalSolvent (shObject *can)
{
    can->setIlkKnown ();
    I->p ("This burns!");
    if (Hero.sufferDamage (UniversalSolventDamage)) {
        Hero.die (kKilled, "drinking a canister of universal solvent");
    }
    Hero.abortion ();
    return FULLTURN;
}

static int
useUniversalSolvent (shObject *can)
{
    int solved = 0;
    can->setIlkKnown ();
    if (Hero.getStoryFlag ("superglued tongue")) {
        I->p ("You dissolve the super glue.");
        Hero.resetStoryFlag ("superglued tongue");
        Hero.mAbil.setByIndex (kPsi, 4 + Hero.mAbil.getByIndex (kPsi));
        ++solved;
    }
    if (Hero.mWeapon && Hero.mWeapon->isBuggy () &&
        Hero.mWeapon->isBugginessKnown ())
    {
        I->p ("You unweld %s.", YOUR (Hero.mWeapon));
        Hero.mWeapon->setDebugged ();
        if (Hero.mWeapon->sufferDamage (UniversalSolventDamage,
                                        Hero.mX, Hero.mY, 1, 1))
        {
            Hero.usedUpItem (Hero.mWeapon, 1, "dissolve");
            delete Hero.removeOneObjectFromInventory (Hero.mWeapon);
        }
        ++solved;
    }
    if (!solved) {
        I->p ("You dissolve some of your flesh.  Ouch!");
        if (Hero.sufferDamage (UniversalSolventDamage)) {
            Hero.die (kKilled, "bathing in universal solvent");
        }
    }
    return FULLTURN;
}

static int
quaffMutagen (shObject *can)
{
    can->setIlkKnown ();
    I->p ("Ick!  That must have been toxic!");
    Hero.getMutantPower ();
    Hero.mRad += NDX (3, 100);
    return FULLTURN;
}

static int
quaffWater (shObject *can)
{
    can->setIlkKnown ();
    I->p ("This tastes like water.");
    return FULLTURN;
}

int
shCreature::healing (int hpmaxdice)
{
    int id = 0;
    if (isHero () && mHP < mMaxHP) {
        I->p ("Your wounds are rapidly healing!");
        id++;
    }
    if (isHero () && Hero.getStoryFlag ("brain incision")) {
        Hero.resetStoryFlag ("brain incision");
        I->p ("Your head wound is closed.");
        id++;
    }
    mHP += NDX (4, 8);
    if (mHP > mMaxHP) {
        mMaxHP += NDX (hpmaxdice, 3);
        mHP = mMaxHP;
        if (isHero () && hpmaxdice) {
            I->p ("You feel much healthier.");
            id++;
        }
    }
    if (isHero () && is (kPlagued)) {
        I->p ("You have a feeling this will delay the plague somewhat.");
        mConditions &= ~kPlagued; /* Seemingly toggle it off. */
        setTimeOut (PLAGUED, 40000, 0); /* Push next attack into future. */
        ++id;
    }
    if (is (kSickened) && !sewerSmells ()) {
        cure (kSickened);
        ++id;
    }
    static const shCondition remedy[] = {kViolated, kConfused, kStunned};
    const int size = sizeof (remedy) / sizeof (shCondition);
    for (int i = 0; i < size; ++i) {
        if (is (remedy[i])) {
            cure (remedy[i]);
            ++id;
        }
    }
    checkTimeOuts ();
    return id;
}

static int
quaffHealing (shObject *can)
{
    if (Hero.healing (1))
        can->setIlkKnown ();
    return FULLTURN;
}

int
shCreature::fullHealing (int hpmaxdice)
{
    int id = 0;
    if (isHero () && mHP < mMaxHP) {
        I->p ("Your wounds are fully healed!");
        id++;
    }
    if (isHero () && Hero.getStoryFlag ("brain incision")) {
        Hero.resetStoryFlag ("brain incision");
        I->p ("Your head wound is closed.");
        id++;
    }
    mHP += NDX (4, 8);
    if (mHP > mMaxHP) {
        mMaxHP += NDX (hpmaxdice, 6);
        if (isHero () && hpmaxdice) {
            I->p ("You feel much healthier.");
            id++;
        }
    }
    mHP = mMaxHP;
    static const shCondition remedy[] = {kPlagued, kViolated, kConfused, kStunned};
    const int size = sizeof (remedy) / sizeof (shCondition);
    for (int i = 0; i < size; ++i) {
        if (is (remedy[i])) {
            cure (remedy[i]);
            ++id;
        }
    }
    cure (kPlagued); /* Plague might be stalled by weak healing. */
    checkTimeOuts ();
    return id;
}

static int
quaffFullHealing (shObject *can)
{
    if (Hero.fullHealing (2))
        can->setIlkKnown ();
    return FULLTURN;
}

static int
quaffRestoration (shObject *can)
{
    if (Hero.restoration ()) {
        can->setIlkKnown ();
    } else {
        can->maybeName ();
    }
    return FULLTURN;
}

static int
quaffBrain (shObject *can)
{
    can->setIlkKnown ();

    I->p ("Brain food!");
    if (!can->isBuggy () && Hero.mMaxAbil.mInt < 25) {
        ++Hero.mMaxAbil.mInt;
    }
    if (Hero.mAbil.mInt < Hero.mMaxAbil.mInt) {
        ++Hero.mAbil.mInt;
    }
    Hero.computeIntrinsics ();
    return FULLTURN;
}

int
quaffGainAbility (shObject *can)
{
    shAbilityIndex permute[6] = {kStr, kCon, kAgi, kDex, kInt, kPsi};

    if (can) {
        can->setIlkKnown ();
    }
    shuffle (permute, 6, sizeof (shAbilityIndex));
    for (int i = 0; i < ((can && can->isOptimized ()) ? 6 : 1); ++i) {
        int a = Hero.mAbil.getByIndex (permute[i]);
        int m = Hero.mMaxAbil.getByIndex (permute[i]);
        if (m - Hero.mExtraAbil.getByIndex (permute[i]) < 25) {
            I->setColor (kBrightGreen);
            I->p ("You feel %s",
                  permute[i] == kStr ? "strong! (+1 Str)" :
                  permute[i] == kCon ? "tough! (+1 Con)" :
                  permute[i] == kAgi ? "agile! (+1 Agi)" :
                  permute[i] == kDex ? "deft! (+1 Dex)" :
                  permute[i] == kInt ? "smart! (+1 Int)" :
                  permute[i] == kPsi ? "weird! (+1 Psi)" : "bugged!");
            if (permute[i] == kCon) {
                if (ABILITY_MODIFIER (m) != ABILITY_MODIFIER (m + 1)) {
                    Hero.mHP += Hero.mCLevel;
                    Hero.mMaxHP += Hero.mCLevel;
                }
            }
            Hero.mAbil.setByIndex (permute[i], a + 1);
            Hero.mMaxAbil.setByIndex (permute[i], m + 1);
            I->setColor (kGray);
        } else {
            I->p ("You're already as %s as you can get!",
                  permute[i] == kStr ? "strong" :
                  permute[i] == kCon ? "tough" :
                  permute[i] == kAgi ? "agile" :
                  permute[i] == kDex ? "dexterous" :
                  permute[i] == kInt ? "smart" :
                  permute[i] == kPsi ? "weird" : "bugged");
        }
    }
    Hero.computeIntrinsics ();
    return FULLTURN;
}

static int
quaffRadAway (shObject *can)
{
    if (Hero.mRad > 0) {
        Hero.mRad -= RNG (100, 300);
        if (Hero.mRad < 0) {
            Hero.mRad = 0;
        }
    }
    if (!Hero.mRad)
        I->p ("You feel purified.");
    else
        I->p ("You feel less contaminated.");

    can->setIlkKnown ();
    return FULLTURN;
}

static int
quaffSpeed (shObject *can)
{
    int numdice = 12;
    if (can->isOptimized ()) {
        numdice = 20;
    } else if (can->isBuggy ()) {
        numdice = 4;
    }
    if (Hero.is (kHosed)) {
        Hero.cure (kHosed);
        Hero.checkTimeOuts ();
    }
    Hero.inflict (kSpeedy, FULLTURN * NDX (numdice, 20));
    I->p ("You feel speedy!");
    can->setIlkKnown ();
    return FULLTURN;
}

static int
quaffPlasma (shObject *can)
{
    can->setIlkKnown ();
    I->p ("It shocks you!");
    if (Hero.sufferDamage (PlasmaDamage)) {
        Hero.die (kKilled, "drinking a canister of plasma");
    }
    Hero.abortion ();
    return FULLTURN;
}

static int
quaffNapalm (shObject *can)
{
    can->setIlkKnown ();
    I->p ("It ignites!");
    if (Hero.sufferDamage (NapalmDamage)) {
        Hero.die (kKilled, "drinking a canister of napalm");
    }
    Hero.abortion ();
    return FULLTURN;
}

static int
quaffLNO (shObject *can)
{
    can->setIlkKnown ();
    I->p ("It freezes!");
    if (Hero.sufferDamage (LiquidNitrogenDamage)) {
        Hero.die (kKilled, "drinking a canister of liquid nitrogen");
    }
    Hero.abortion ();
    return FULLTURN;
}

static int
quaffPoison (shObject *can)
{
    if (Hero.mBodyArmor && Hero.mBodyArmor->isA (kObjGreenCPA)) {
        I->p ("This nourishing drink has delicious taste!");
        Hero.healing (0);
        int con = Hero.mAbil.mCon;
        int maxcon = Hero.mMaxAbil.mCon;
        if (con < maxcon) {
            if (can->isOptimized ()) {
                Hero.mAbil.mCon = maxcon;
                con = maxcon;
            } else {
                ++Hero.mAbil.mCon;
                ++con;
            }
            if (con == maxcon) {
                I->p ("You feel healthy.");
            } else {
                I->p ("You feel healthier.");
            }
        }
        can->maybeName ();
    } else {
        can->setIlkKnown ();
        I->p ("This stuff must be poisonous.");
        if (Hero.sufferDamage (PoisonDamage)) {
            Hero.die (kKilled, "drinking a canister of poison");
        }
        Hero.abortion ();
    }
    return FULLTURN;
}

static int
quaffSpice (shObject *can)
{
    can->setIlkKnown ();
    I->p ("You fold space.");
    if (1 == Hero.transport (-1, -1, 100, 1)) {
        Hero.die (kKilled, "folding space");
    }
    I->p ("Your senses open to new dimensions.");
    Hero.computeIntrinsics ();
    Hero.setTimeOut (TELEPATHY, FULLTURN * NDX (10, 30));
    Hero.setTimeOut (XRAYVISION, FULLTURN * NDX (10, 30));
    Hero.inflict (kSpeedy, FULLTURN * NDX (5, 2));
    Hero.mInnateIntrinsics |= kTelepathy | kXRayVision;
    /* Add +20% chances for each Voice effect in power over you. */
    int chance = (Hero.is (kUnableCompute) > 0) * 20 +
        (Hero.is (kUnableUsePsi) > 0) * 20 + (Hero.is (kUnableAttack) > 0) * 20;
    if (RNG (100) < chance) {
        Hero.getMutantPower (kTheVoice);
    }
    return FULLTURN;
}

static int
quaffEmbryo (shObject *can)
{
    can->setIlkKnown ();
	I->p ("You drink something writhing and slimy.");
	Hero.setStoryFlag ("impregnation", 1);
	return FULLTURN;
}

static int
quaffB3 (shObject *can)
{
    can->setIlkKnown ();
    I->p ("You feel happier!");
    Hero.mRad += RNG (1, 10);
    if (RNG (2) && Hero.mHP >= 2) {
        --Hero.mHP;
        I->drawSideWin ();
    }
    if (!RNG (3)) {
        Hero.abortion ();
    }
    if (!RNG (40)) {
        I->p ("Your health suffers.");
        Hero.sufferAbilityDamage (kCon, 1);
        I->drawSideWin ();
    }
    if (!RNG (100)) {
        Hero.getMutantPower ();
    }
    int amt = RNG (1, 5);
    Hero.mPsiDrain -= amt;
    Hero.mAbil.mPsi += amt;
    return FULLTURN;
}

static int
useCanister (shObject *can)
{
    int t = useCommonCanister (can);
    if (t) {
        if (Hero.isInShop ()) {
            Hero.usedUpItem (can, 1, "use");
        }
        Hero.useUpOneObjectFromInventory (can);
    }
    return t;
}


void
initializeCanisters ()
{
    int n = 0;

    /* enum: kObjGenericCanister */
    /* enum: kObjFirstCanister = kObjGenericCanister */
    NEW_CANISTER(kObjGenericCanister,
        "canister", "canister",
        kWhite, 0,
        NULL, 0 BZ, ABSTRACT);
    AllIlks[kObjGenericCanister]->mParent = kObjNothing;

    /* enum: kObjBeer */
    NEW_CANISTER(kObjBeer,
        "canister of beer",
        CanisterData[n].mDesc, CanisterData[n++].mColor, 0,
        &quaffBeer, 5 BZ, 100);

    /* enum: kObjSuperGlue */
    NEW_EXPLOSIVE_CANISTER(kObjSuperGlue,
        "canister of super glue",
        CanisterData[n].mDesc, CanisterData[n++].mColor, 0,
        newGrenadeAttack (shAttack::kBlast, 1, kParalyzing, 2, 5),
        &quaffSuperGlue, 5 BZ, 50);
    /* TODO: kGluing damage type. */

    /* enum: kObjNanoCola */
    NEW_CANISTER(kObjNanoCola,
        "canister of nano cola",
        CanisterData[n].mDesc, CanisterData[n++].mColor, 0,
        &quaffNanoCola, 5 BZ, 100);

    /* enum: kObjCoffee */
    NEW_CANISTER(kObjCoffee,
        "canister of coffee",
        CanisterData[n].mDesc, CanisterData[n++].mColor, 0,
        &quaffCoffee, 5 BZ, 50);

    /* enum: kObjWater */
    NEW_CANISTER(kObjWater,
        "canister of water",
        "clear canister", kBrightCyan, kPreidentified,
        &quaffWater, 5 BZ, 10);

    /* enum: kObjB3 */
    NEW_EXPLOSIVE_CANISTER(kObjB3,
        "canister of Bouncy Bubbly Beverage",
        CanisterData[n].mDesc, CanisterData[n++].mColor, 0,
        newGrenadeAttack (shAttack::kBlast, 0, kConcussive, 3, 9),
        &quaffB3, 5 BZ, 25);


    /* No doozies in the 50 buckazoid category. */
    /* enum: kObjRadAway */
    NEW_CANISTER(kObjRadAway,
        "canister of Rad-Away",
        CanisterData[n].mDesc, CanisterData[n++].mColor, 0,
        &quaffRadAway, 50 BZ, 75);

    /* enum: kObjRestorationCanister */
    NEW_CANISTER(kObjRestorationCanister,
        "canister of restoration",
        CanisterData[n].mDesc, CanisterData[n++].mColor, 0,
        &quaffRestoration, 50 BZ, 50);

    /* enum: kObjHealingCanister */
    NEW_CANISTER(kObjHealingCanister,
        "canister of healing",
        CanisterData[n].mDesc, CanisterData[n++].mColor, 0,
        &quaffHealing, 50 BZ, 125);


    /* enum: kObjLNO */
    NEW_EXPLOSIVE_CANISTER(kObjLNO,
        "canister of liquid nitrogen",
        CanisterData[n].mDesc, CanisterData[n++].mColor, 0,
        newGrenadeAttack (shAttack::kBlast, 1, kFreezing, 4, 32),
        &quaffLNO, 75 BZ, 50);

    /* enum: kObjNapalm */
    NEW_EXPLOSIVE_CANISTER(kObjNapalm,
        "canister of napalm",
        CanisterData[n].mDesc, CanisterData[n++].mColor, 0,
        newGrenadeAttack (shAttack::kBlast, 1, kBurning, 4, 32),
        &quaffNapalm, 75 BZ, 50);


    /* enum: kObjUniversalSolvent */
    NEW_EXPLOSIVE_CANISTER(kObjUniversalSolvent,
        "canister of universal solvent",
        CanisterData[n].mDesc, CanisterData[n++].mColor, 0,
        newGrenadeAttack (shAttack::kBlast, 1, kCorrosive, 6, 14),
        &quaffUniversalSolvent, 100 BZ, 25);

    /* enum: kObjSpeedCanister */
    NEW_CANISTER(kObjSpeedCanister,
        "canister of speed",
        CanisterData[n].mDesc, CanisterData[n++].mColor, 0,
        &quaffSpeed, 100 BZ, 50);

    /* enum: kObjPoisonCanister */
    NEW_CANISTER(kObjPoisonCanister,
        "canister of poison",
        CanisterData[n].mDesc, CanisterData[n++].mColor, 0,
        &quaffPoison, 100 BZ, 50);

    /* enum: kObjPlasmaCanister */
    NEW_EXPLOSIVE_CANISTER(kObjPlasmaCanister,
        "canister of plasma",
        CanisterData[n].mDesc, CanisterData[n++].mColor, 0,
        newGrenadeAttack (shAttack::kBlast, 1,
            kBurning, 4, 32, kMagnetic, 4, 8),
        &quaffPlasma, 100 BZ, 50);


    /* These are radioactive unless optimized. */
    /* enum: kObjMutagenCanister */
    NEW_CANISTER(kObjMutagenCanister,
        "canister of mutagen",
        CanisterData[n].mDesc, CanisterData[n++].mColor, 0,
        &quaffMutagen, 200 BZ, 50);

    /* enum: kObjFullHealingCanister */
    NEW_CANISTER(kObjFullHealingCanister,
        "canister of full healing",
        CanisterData[n].mDesc, CanisterData[n++].mColor, 0,
        &quaffFullHealing, 200 BZ, 50);

    /* enum: kObjGainAbilityCanister */
    NEW_CANISTER(kObjGainAbilityCanister,
        "canister of gain ability",
        CanisterData[n].mDesc, CanisterData[n++].mColor, 0,
        &quaffGainAbility, 200 BZ, 25);

    /* enum: kObjSpiceMelange */
    NEW_CANISTER(kObjSpiceMelange,
        "canister of spice",
        CanisterData[n].mDesc, CanisterData[n++].mColor, 0,
        &quaffSpice, 200 BZ, 50);


    /* enum: kObjAntimatter */
    NEW_EXPLOSIVE_CANISTER(kObjAntimatter,
        "canister of antimatter",
        "canister with death symbol", kGray, 0,
        newGrenadeAttack (shAttack::kBlast, 2,
            kDisintegrating, 20, 40, kConcussive, 4, 16),
        &quaffAntimatter, 1000 BZ, 10);

    /* enum: kObjBrain */
    NEW_CANISTER(kObjBrain,
        "brain cylinder",
        "brain in a jar", kBrown, kPreidentified,
        &quaffBrain, 1000 BZ, 10);

    /* enum: kObjCannedEmbryo */
    /* enum: kObjLastCanister = kObjCannedEmbryo */
    NEW_CANISTER(kObjCannedEmbryo,
        "canned alien embryo",
        "veiny canister", kBrown, 0,
        &quaffEmbryo, 1000 BZ, 5);

}


int
quaffCanister (shObject *can)
{
    int t;

    assert (can->isA (kCanister));

    if (Hero.isInShop ()) {
        Hero.usedUpItem (can, 1, "drink");
        can->resetUnpaid ();
    }
    t = (can->myIlk ()->mQuaffFunc) (can);
    if (can->isIlkKnown ()) { /* Identify rest of stack. */
        can->setAppearanceKnown ();
    }
    Hero.useUpOneObjectFromInventory (can);
    return t;
}
